#!/usr/bin/env python

import subprocess, os, shutil, argparse, struct
import py_wiivc_disable_gxsetcopyfilter_util as gxsetcopyfilter_util

def setup_environment(bins: list = ["wadunpacker","wadpacker","gbalzss"], path: str = os.path.realpath(".")+"/bin") -> None:
    for b in bins:
        f = path+"/"+b
        #ensure exec perms
        if (os.stat(f).st_mode & 0o777) != 0o755: os.chmod(f,0o755);
    os.environ.update({"PATH":os.environ["PATH"]+":"+path})

def create_tempdir(path: str = "/tmp/wad") -> str:
    if os.path.isdir(path) is False:
        try:
            os.mkdir(path)
        except FileExistsError:
            print("%s already exists" %(path))
        return(path)

def unpack_wad(fn: str, wadpath: str) -> str:
    dest = shutil.copy(fn,wadpath)
    destdir = dest[0:dest.rindex("/")]
    os.chdir(destdir)
    with gxsetcopyfilter_util.Shell(["wadunpacker",fn]) as sh:
        sh.run()
        res = sh.get_result()
    if res.returncode == 0:
        return(destdir)
    else:
        print("unpacking of %s failed" %(os.path.realpath(fn)))
        exit(res.returncode)

def check_lz77(f: str,offset: int) -> bool:
    # code from Marcan (https://wiibrew.org/wiki/LZ77)
    TYPE_LZ77 = 1
    f.seek(offset)
    hdr = struct.unpack("<I",f.read(4))[0]
    uncompressed_length = hdr>>8
    compression_type = hdr>>4 & 0xF
    if compression_type != TYPE_LZ77:
        return(False)
    else:
        return(True)

def handle_lz77(action: str,fn: str, outfn: str) -> str:
    if action == "decompress":
        shell_cmd = [ "gbalzss","--lz11","d",fn,outfn ]
    elif action == "compress":
        shell_cmd = [ "gbalzss","--lz11","e",fn,outfn ]
    else:
        raise ValueError("%s not a valid action" %(action))
    with gxsetcopyfilter_util.Shell(shell_cmd) as sh:
        sh.run()
        res = sh.get_result()
    if res.returncode == 0:
        os.remove(os.path.realpath(fn))
        return(os.path.realpath(outfn))
    else:
        print("uncompression of %s failed" %(fn))
        exit(res.returncode)

def disable_filter(wadpath: str, fn: str = "00000001.app") -> str:
    # TODO: needs improvment
    os.chdir(wadpath)
    doldir = [ d for top,dirs,files in os.walk(".") for d in dirs ]
    if len(doldir) != 1:
        raise ValueError("one directory is expected, not %i" %(len(doldir)))
    os.chdir(doldir[0])
    #handle compression
    with open(fn,"rb") as f:
        is_lz77 = check_lz77(f,f.tell())
    if is_lz77:
        outfn = fn.rstrip(".app")+"_decompressed.app"
        orgfn = fn
        handle_lz77("decompress",fn,outfn)
        fn = outfn
        
    if gxsetcopyfilter_util.check_filter(fn)[0] == "set":
        gxsetcopyfilter_util.change_filter(fn,"disable")
    else:
        print("already disabled")
        
    if is_lz77:
        handle_lz77("compress",fn,orgfn)
    return(doldir[0])

def pack_wad(wadpath: str, doldir: str, newfn: str, outdir=None) -> str:
    if "/" in newfn:
        newfn = newfn.split("/")[-1] # use name after last /
    dest = wadpath+"/"+newfn
    os.chdir(wadpath)
    with gxsetcopyfilter_util.Shell(["wadpacker",doldir,dest]) as sh:
        sh.run()
        res = sh.get_result()
    if res.returncode == 0:
        if outdir is not None:
            dest = shutil.move(newfn,outdir)
        return(dest)
    else:
        print("packing of %s failed" %(newfn))
        exit(res.returncode)

def cleanup(doldir: str, fn: str) -> None:
    shutil.rmtree(doldir)
    if os.path.exists(fn): os.remove(fn)

def create_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input",help="Input file (required)",required=True)
    parser.add_argument("-o","--output-folder",help="Output folder")
    parser.add_argument("--new-filename",help="New filename")
    parser.add_argument("-f","--filter_file",help="Filename of the dol which contains GXSetCopyFilter function (default: 00000001.app)",default="00000001.app")
    parser.add_argument("--tmp_waddir",default="/tmp/wad",help="Temp dir to use for wads (default: /tmp/wad)")
    parser.add_argument("--bindir",default=os.path.realpath(".")+"/bin",help="Directory which contains needed binaries (default: "+os.path.realpath(".")+"/bin")
    return(parser.parse_args())

if __name__ == "__main__":
    args = create_parser()
    newfn = args.input.rstrip(".wad")+"(No VFilter).wad" if args.new_filename is None else args.new_filename
    setup_environment(path=args.bindir)
    create_tempdir(args.tmp_waddir)
    destdir = unpack_wad(args.input,args.tmp_waddir)
    doldir = disable_filter(destdir,args.filter_file)
    if args.output_folder is not None:
        out = pack_wad(destdir,doldir,newfn,args.output_folder)
    else:
        out = pack_wad(destdir,doldir,newfn)
    print("new wad in %s" %(out))
    cleanup(doldir,args.input[args.input.rindex("/")+1:])
